//
// Copyright (c) 2017 SWET. All rights reserved.
//

import Foundation
import Firebase

extension Auth {

    enum OperationType {
        case login
        case register
    }

    /// Firebaseのエラーをローカライズしたメッセージに変換する
    ///
    /// - Parameters:
    ///   - error: Firebaseから返却されたエラー
    ///   - operationType: オペレーション種別
    /// - Returns: ローカライズされたエラーメッセージ
    static func localizedErrorMessage(from error: Error, for operationType: OperationType) -> String {
        
        typealias ErrorArgs = (key: String, comment: String)

        let args: ErrorArgs

        let loginFailedArgs    = ErrorArgs(key: "error_login_failed", comment: "ログインに失敗しました。")
        let registerFailedArgs = ErrorArgs(key: "error_register_failed", comment: "ユーザ登録に失敗しました。")
        let invalidEmailArgs   = ErrorArgs(key: "error_invalidEmail", comment: "メールアドレスの形式が正しくありません。")

        switch operationType {
        case .login:
            guard let errorCode = AuthErrorCode(rawValue: error._code) else {
                return NSLocalizedString(loginFailedArgs.key, comment: loginFailedArgs.comment)
            }

            switch errorCode {
            case .invalidEmail:
                args = invalidEmailArgs
            case .userDisabled:
                args = ErrorArgs(key: "error_login_userDisabled", comment: "アカウントが無効です。")
            case .wrongPassword:
                args = ErrorArgs(key: "error_login_wrongPassword", comment: "メールアドレスまたはパスワードが間違っています。")
            case .operationNotAllowed: fallthrough
            default:
                args = loginFailedArgs
            }

        case .register:
            guard let errorCode = AuthErrorCode(rawValue: error._code) else {
                return NSLocalizedString(registerFailedArgs.key, comment: registerFailedArgs.comment)
            }

            switch errorCode {
            case .invalidEmail:
                args = invalidEmailArgs
            case .emailAlreadyInUse:
                args = ErrorArgs(key: "error_used_email", comment: "既にメールアドレスは使用されています。")
            case .weakPassword:
                args = ErrorArgs(key: "error_register_weakPassword", comment: "パスワードが弱すぎます。")
            case .operationNotAllowed: fallthrough
            default:
                args = registerFailedArgs
            }
        }

        return NSLocalizedString(args.key, comment: args.comment)
    }

}
